##
# File:    ProvenanceProvider.py
# Author:  J. Westbrook
# Date:    25-Aug-2019
# Version: 0.001
#
# Updates:
#
##
"""
Utilities to access and update provenance details.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class ProvenanceProvider(SingletonClass):
    """ Utilities to access and update provenance details.
    """

    def __init__(self, cfgOb, cachePath, useCache=True, **kwargs):
        """Utilities to access and update provenance details.

        Args:
            cfgOb ([type]): ConfigInfo() instance
            cachePath ([type]): path to directory containing schema
            useCache (bool, optional): use cached schema. Defaults to True.
        """

        self.__cfgOb = cfgOb
        self.__configName = self.__cfgOb.getDefaultSectionName()
        self.__cachePath = cachePath
        self.__useCache = useCache
        #
        self.__workPath = os.path.join(self.__cachePath, "work")
        self.__provenanceCachePath = os.path.join(self.__cachePath, self.__cfgOb.get("PROVENANCE_INFO_CACHE_DIR", sectionName=self.__configName))
        self.__provenanceLocator = self.__cfgOb.getPath("PROVENANCE_INFO_LOCATOR", sectionName=self.__configName)
        #
        self.__fileU = FileUtil(workPath=self.__workPath)
        self.__fileU.mkdir(self.__provenanceCachePath)
        self.__kwargs = kwargs
        #

    def __reload(self, locator, dirPath, useCache=True):
        #
        fn = self.__fileU.getFileName(locator)
        filePath = os.path.join(dirPath, fn)
        logger.debug("Using cache path %s", dirPath)
        self.__fileU.mkdir(dirPath)
        if not useCache:
            try:
                os.remove(filePath)
            except Exception:
                pass
        #
        if useCache and self.__fileU.exists(filePath):
            ok = True
        else:
            logger.debug("Fetch data from source %s", locator)
            ok = self.__fileU.get(locator, filePath)

        return filePath if ok else None

    def fetch(self):
        try:
            provenanceFileCachePath = self.__reload(self.__provenanceLocator, self.__provenanceCachePath, useCache=self.__useCache)
            mU = MarshalUtil(workPath=self.__workPath)
            return mU.doImport(provenanceFileCachePath, fmt="json")
        except Exception as e:
            logger.exception("Failed retreiving provenance with %s", str(e))
        return {}

    def update(self, provD):
        ok = False
        try:
            provenanceFileCachePath = self.__reload(self.__provenanceLocator, self.__provenanceCachePath, useCache=self.__useCache)
            mU = MarshalUtil(workPath=self.__workPath)
            tD = mU.doImport(provenanceFileCachePath, fmt="json")
            tD.update(provD)
            ok = mU.doExport(provenanceFileCachePath, tD, fmt="json")
        except Exception as e:
            logger.exception("Failed updating provenance with %s", str(e))
        return ok

    def store(self, provD):
        ok = False
        try:
            provenanceFileCachePath = self.__reload(self.__provenanceLocator, self.__provenanceCachePath, useCache=self.__useCache)
            mU = MarshalUtil(workPath=self.__workPath)
            ok = mU.doExport(provenanceFileCachePath, provD, fmt="json")
        except Exception as e:
            logger.exception("Failed storing provenance with %s", str(e))
        return ok



##
# File   makePathList.py
# Date:  18-Feb-2018
#
#  Get the current list of release entry files in our data release file system.
#
##
import logging
import os

try:
    import os.scandir as scandir
except Exception:
    import scandir

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def makePdbxPathList(fp, cachePath=".", skipFile=None):
    """ Return the list of pdbx file paths in the current repository.
    """

    try:
        skipD = {}
        if skipFile and os.access(skipFile, "r"):
            with open(skipFile, "r") as ifh:
                for line in ifh:
                    idcode = str(line[:-1]).strip().lower() + ".cif"
                    skipD[idcode] = idcode
            logger.info("Skip list length %d", len(skipD))
        #
        with open(fp, "w") as ofh:
            for root, _, files in scandir.walk(cachePath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.endswith(".cif") and len(name) == 8 and name not in skipD:
                        ofh.write("%s\n" % os.path.join(root, name))
            #
            # logger.info("\nFound %d files in %s\n" % (len(pathList), cachePath))
        return True
    except Exception as e:
        logger.exception("Failing with %s", str(e))

    return False


if __name__ == "__main__":
    ok = makePdbxPathList("PDBXPATHLIST.txt", cachePath="/net/beta_data/mmcif-pdbx-load-v5.0", skipFile="./DONE.LIST")



##
# File:    SchemaDefBuild.py
# Author:  J. Westbrook
# Date:    1-May-2018
# Version: 0.001 Initial version
#
# Updates:
#
#  9-May-2018 jdw integrate dictionary and file based (type/coverage) data.
#  7-Aug-2018 jdw add slice definitions converted to schema id references
# 13-Aug-2018 jdw Refine the role of includeContentClasses -
# 14-Aug-2018 jdw Return 'COLLECTION_DOCUMENT_ATTRIBUTE_NAMES' as a list
#  6-Sep-2018 jdw Generalize JSON schema generation method
# 14-Sep-2018 jdw Require at least one record in any array type, adjust constraints on iterables.
# 18-Sep-2018 jdw Constrain categories/class to homogeneous content
#  7-Oct-2018 jdw Add subCategory aggregation in the JSON schema generator
#  9-Oct-2018 jdw push the constructor arguments into the constructor as configuration options
# 12-Oct-2018 jdw filter empty required attributes in subcategory aggregates
# 24-Oct-2018 jdw update for new configuration organization
# 18-Nov-2018 jdw add COLLECTION_DOCUMENT_ATTRIBUTE_INFO
#  3-Dec-2018 jdw add INTEGRATED_CONTENT
#  6-Jan-2019 jdw update to the change in configuration for dataTypeInstanceFile
# 16-Jan-2019 jdw add 'COLLECTION_DOCUMENT_REPLACE_ATTRIBUTE_NAMES'
# 31-Mar-2019 jdw add  support for 'addParentRefs' in enforceOpts to include relative $ref properties
#                 to describe parent relationships
#  3-Apr-2019 jdw add experimental primary key property controlled by 'addPrimaryKey'
# 22-Aug-2019 jdw DictInfo() replaced with new ContentInfo()
##
"""
Integrate dictionary metadata and file based (type/coverage) into internal and JSON/BSON schema defintions.


"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

# pylint: disable=too-many-lines

import copy
import logging
from collections import OrderedDict

from rcsb.db.define.DataTypeApiProvider import DataTypeApiProvider
from rcsb.db.define.ContentDefinition import ContentDefinition
from rcsb.db.define.DictionaryApiProviderWrapper import DictionaryApiProviderWrapper

logger = logging.getLogger(__name__)


class SchemaDefBuild(object):
    """ Integrate dictionary metadata and file based(type/coverage) into internal and JSON/BSON schema defintions.
    """

    def __init__(self, databaseName, cfgOb, cachePath=None, includeContentClasses=None):
        """Integrate dictionary metadata and file based(type/coverage) into internal and JSON/BSON schema defintions.

        Args:
            databaseName (str): schema name
            cfgOb (object): ConfigInfo() object instance
            cachePath (str): path to cached resources
            includeContentClasses (list, optional): content class list. Defaults to None.
        """
        configName = "site_info_configuration"
        self.__cfgOb = cfgOb
        self.__databaseName = databaseName
        self.__cachePath = cachePath if cachePath else "."
        self.__includeContentClasses = (
            includeContentClasses if includeContentClasses else ["GENERATED_CONTENT", "EVOLVING_CONTENT", "CONSOLIDATED_BIRD_CONTENT", "INTEGRATED_CONTENT"]
        )
        #
        self.__contentDefHelper = self.__cfgOb.getHelper("CONTENT_DEF_HELPER_MODULE", sectionName=configName, cfgOb=self.__cfgOb)
        self.__documentDefHelper = self.__cfgOb.getHelper("DOCUMENT_DEF_HELPER_MODULE", sectionName=configName, cfgOb=self.__cfgOb)
        #
        self.__dtP = DataTypeApiProvider(self.__cfgOb, cachePath, useCache=True)
        #
        dP = DictionaryApiProviderWrapper(cfgOb, self.__cachePath, useCache=True)
        dictApi = dP.getApiByName(databaseName)
        self.__contentInfo = ContentDefinition(dictApi, contentDefHelper=self.__contentDefHelper, databaseName=databaseName)
        #

    def build(self, collectionName=None, dataTyping="ANY", encodingType="rcsb", enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums"):
        rD = {}
        if encodingType.lower() == "rcsb":
            rD = self.__build(
                databaseName=self.__databaseName,
                dataTyping=dataTyping,
                contentDefHelper=self.__contentDefHelper,
                documentDefHelper=self.__documentDefHelper,
                includeContentClasses=self.__includeContentClasses,
            )
        elif encodingType.lower() in ["json", "bson"]:
            rD = self.__createJsonLikeSchema(
                databaseName=self.__databaseName,
                collectionName=collectionName,
                dataTyping=dataTyping.upper(),
                contentDefHelper=self.__contentDefHelper,
                documentDefHelper=self.__documentDefHelper,
                includeContentClasses=self.__includeContentClasses,
                enforceOpts=enforceOpts,
            )
        return rD

    def __build(self, databaseName, dataTyping, contentDefHelper, documentDefHelper, includeContentClasses):
        """
        """
        databaseVersion = contentDefHelper.getDatabaseVersion(databaseName) if contentDefHelper else ""

        #
        schemaDef = {"NAME": databaseName, "APP_NAME": dataTyping, "DATABASE_NAME": databaseName, "DATABASE_VERSION": databaseVersion}
        #
        schemaDef["SELECTION_FILTERS"] = self.__contentInfo.getSelectionFiltersForDatabase()

        schemaDef["SCHEMA_DICT"] = self.__createSchemaDict(databaseName, dataTyping, contentDefHelper, includeContentClasses)
        schemaDef["DOCUMENT_DICT"] = self.__createDocumentDict(databaseName, documentDefHelper)
        schemaDef["SLICE_PARENT_ITEMS"] = self.__convertSliceParentItemNames(databaseName, dataTyping)
        schemaDef["SLICE_PARENT_FILTERS"] = self.__convertSliceParentFilterNames(databaseName, dataTyping)
        return schemaDef

    def __createDocumentDict(self, databaseName, documentDefHelper):
        """Internal method to assign document-level details to the schema definition,


        Args:
            databaseName (string): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            documentDefHelper (class instance):  Class instance providing additional document-level metadata

        Returns:
            dict: dictionary of document-level metadata


        """
        rD = {
            "CONTENT_TYPE_COLLECTION_INFO": [],
            "COLLECTION_DOCUMENT_ATTRIBUTE_NAMES": {},
            "COLLECTION_DOCUMENT_REPLACE_ATTRIBUTE_NAMES": {},
            "COLLECTION_DOCUMENT_PRIVATE_KEYS": {},
            "COLLECTION_DOCUMENT_INDICES": {},
            "COLLECTION_CONTENT": {},
            "COLLECTION_SUB_CATEGORY_AGGREGATES": {},
        }
        #
        dH = documentDefHelper
        if dH:
            # cdL  = list of [{'NAME': , 'VERSION': xx }, ...]
            cdL = dH.getCollectionInfo(databaseName)
            rD["CONTENT_TYPE_COLLECTION_INFO"] = cdL
            for cd in cdL:
                cN = cd["NAME"]
                rD["COLLECTION_CONTENT"][cN] = {
                    "INCLUDE": dH.getIncluded(cN),
                    "EXCLUDE": dH.getExcluded(cN),
                    "SLICE_FILTER": dH.getSliceFilter(cN),
                    "EXCLUDED_ATTRIBUTES": dH.getDocumentExcludedAttributes(cN, asTuple=False),
                }
                rD["COLLECTION_DOCUMENT_ATTRIBUTE_NAMES"][cN] = dH.getDocumentKeyAttributeNames(cN)
                rD["COLLECTION_DOCUMENT_REPLACE_ATTRIBUTE_NAMES"][cN] = dH.getDocumentReplaceAttributeNames(cN)
                rD["COLLECTION_DOCUMENT_PRIVATE_KEYS"][cN] = dH.getPrivateDocumentAttributes(cN)
                rD["COLLECTION_DOCUMENT_INDICES"][cN] = dH.getDocumentIndices(cN)
                rD["COLLECTION_SUB_CATEGORY_AGGREGATES"][cN] = dH.getSubCategoryAggregateFeatures(cN)
        #
        return rD

    def __testContentClasses(self, includeContentClasses, assignedContentClasses):
        """ Return True if any of the include content classes are assigned.

        """
        # logger.debug("includeContentClasses %r assignedContentClasses %r" % (includeContentClasses, assignedContentClasses))
        for cc in includeContentClasses:
            if cc in assignedContentClasses:
                return True
        return False

    def __getConvertNameMethod(self, dataTyping):
        # Function to perform category and attribute name conversion.
        # convertNameF = self.__contentDefHelper.convertNameDefault if self.__contentDefHelper else self.__convertNameDefault
        #
        try:
            if dataTyping in ["ANY", "SQL", "DOCUMENT", "SOLR", "JSON", "BSON"]:
                nameConvention = dataTyping
            else:
                nameConvention = "DEFAULT"
            return self.__contentDefHelper.getConvertNameMethod(nameConvention) if self.__contentDefHelper else self.__convertNameDefault
        except Exception:
            pass

        return self.__convertNameDefault

    def __createSchemaDict(self, databaseName, dataTyping, contentDefHelper, includeContentClasses=None):
        """Internal method to integrate dictionary and instance metadata into a common schema description data structure.

        Args:
            databaseName (string): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            dataTyping (string): ANY|SQL
            contentDefHelper (class instance): Class instance providing additional schema details
            includeContentClasses (list, optional): list of additional content classes to be included (e.g. GENERATED_CONTENT)

        Returns:
            dict: definitions for each schema object


        """
        verbose = False
        contentClasses = includeContentClasses if includeContentClasses else []
        #
        logger.debug("Including additional category classes %r", contentClasses)
        #
        dtInstInfo = self.__dtP.getDataTypeInstanceApi(databaseName)
        dtAppInfo = self.__dtP.getDataTypeApplicationApi(dataTyping)
        #
        # Supplied by the contentDefHelper
        #
        includeList = contentDefHelper.getIncluded(databaseName) if contentDefHelper else []
        excludeList = contentDefHelper.getExcluded(databaseName) if contentDefHelper else []
        excludeAttributesD = contentDefHelper.getExcludedAttributes(databaseName) if contentDefHelper else {}
        #
        logger.debug("Schema include list length %d", len(includeList))
        logger.debug("Schema exclude list length %d", len(excludeList))
        #
        # Optional synthetic attribute added to each category with value linked to data block identifier (or other function)
        #
        blockAttributeName = contentDefHelper.getBlockAttributeName(databaseName) if contentDefHelper else None
        blockAttributeCifType = contentDefHelper.getBlockAttributeCifType(databaseName) if contentDefHelper else None
        blockAttributeAppType = dtAppInfo.getAppTypeName(blockAttributeCifType)
        blockAttributeWidth = contentDefHelper.getBlockAttributeMaxWidth(databaseName) if contentDefHelper else 0
        blockAttributeMethod = contentDefHelper.getBlockAttributeMethod(databaseName) if contentDefHelper else None
        #
        convertNameF = self.__getConvertNameMethod(dataTyping)
        #
        dictSchema = self.__contentInfo.getSchemaNames()
        logger.debug("Dictionary category length %d", len(dictSchema))
        #
        rD = OrderedDict()
        for catName, fullAtNameList in dictSchema.items():

            #
            atNameList = [at for at in fullAtNameList if (catName, at) not in excludeAttributesD]
            #
            cfD = self.__contentInfo.getCategoryFeatures(catName)
            #
            # logger.debug("catName %s contentClasses %r cfD %r" % (catName, contentClasses, cfD))

            if not dtInstInfo.exists(catName) and not self.__testContentClasses(contentClasses, cfD["CONTENT_CLASSES"]):
                logger.debug("Schema %r Skipping category %s content classes %r", databaseName, catName, cfD["CONTENT_CLASSES"])
                continue
            sName = convertNameF(catName)
            sId = sName.upper()
            #
            if excludeList and sId in excludeList:
                continue
            if includeList and sId not in includeList:
                continue
            # JDW
            if not cfD:
                logger.info("%s catName %s contentClasses %r cfD %r", databaseName, catName, contentClasses, cfD)
            #
            aD = self.__contentInfo.getAttributeFeatures(catName)
            #
            sliceNames = self.__contentInfo.getSliceNames()
            dD = OrderedDict()
            dD["SCHEMA_ID"] = sId
            dD["SCHEMA_NAME"] = sName
            dD["SCHEMA_TYPE"] = "transactional"
            dD["SCHEMA_UNIT_CARDINALITY"] = cfD["UNIT_CARDINALITY"] if "UNIT_CARDINALITY" in cfD else False
            dD["SCHEMA_CONTENT_CLASSES"] = cfD["CONTENT_CLASSES"] if "CONTENT_CLASSES" in cfD else []
            dD["SCHEMA_MANDATORY"] = cfD["IS_MANDATORY"]
            dD["SCHEMA_SUB_CATEGORIES"] = []
            #
            dD["ATTRIBUTES"] = OrderedDict({convertNameF(blockAttributeName).upper(): convertNameF(blockAttributeName)}) if blockAttributeName else {}
            #
            dD["ATTRIBUTE_MAP"] = (
                OrderedDict({(convertNameF(blockAttributeName)).upper(): {"CATEGORY": None, "ATTRIBUTE": None, "METHOD_NAME": blockAttributeMethod, "ARGUMENTS": None}})
                if blockAttributeName
                else OrderedDict()
            )

            dD["ATTRIBUTE_INFO"] = OrderedDict()
            atIdIndexList = []
            atNameIndexList = []
            iOrder = 1
            if blockAttributeName:
                td = OrderedDict(
                    {
                        "ORDER": iOrder,
                        "NULLABLE": False,
                        "PRECISION": 0,
                        "PRIMARY_KEY": True,
                        "APP_TYPE": blockAttributeAppType,
                        "WIDTH": blockAttributeWidth,
                        "ITERABLE_DELIMITER": None,
                        "FILTER_TYPES": [],
                        "ENUMERATION": [],
                        "IS_CHAR_TYPE": True,
                        "CONTENT_CLASSES": ["BLOCK_ATTRIBUTE"],
                        "SUB_CATEGORIES": [],
                    }
                )
                iOrder += 1
                atId = (convertNameF(blockAttributeName)).upper()
                atIdIndexList.append(atId)
                atNameIndexList.append(blockAttributeName)
                dD["ATTRIBUTE_INFO"][atId] = td
            #
            for atName in sorted(atNameList):
                fD = aD[atName]
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD["CONTENT_CLASSES"]):
                    continue
                if fD["IS_KEY"]:
                    appType = dtAppInfo.getAppTypeName(fD["TYPE_CODE"])
                    appWidth = dtAppInfo.getAppTypeDefaultWidth(fD["TYPE_CODE"])
                    instWidth = dtInstInfo.getMaxWidth(catName, atName)
                    #
                    revAppType, revAppWidth = dtAppInfo.updateCharType(fD["IS_KEY"], appType, instWidth, appWidth, bufferPercent=20.0)
                    if verbose and dataTyping in ["SQL", "ANY"]:
                        logger.debug(
                            "catName %s atName %s cifType %s appType %s appWidth %r instWidth %r --> revAppType %r revAppWidth %r ",
                            catName,
                            atName,
                            fD["TYPE_CODE"],
                            appType,
                            appWidth,
                            instWidth,
                            revAppType,
                            revAppWidth,
                        )
                    #
                    appPrecision = dtAppInfo.getAppTypeDefaultPrecision(fD["TYPE_CODE"])
                    td = OrderedDict(
                        {
                            "ORDER": iOrder,
                            "NULLABLE": not fD["IS_MANDATORY"],
                            "PRECISION": appPrecision,
                            "PRIMARY_KEY": fD["IS_KEY"],
                            "APP_TYPE": revAppType,
                            "WIDTH": revAppWidth,
                            "ITERABLE_DELIMITER": None,
                            "FILTER_TYPES": fD["FILTER_TYPES"],
                            "IS_CHAR_TYPE": fD["IS_CHAR_TYPE"],
                            "ENUMERATION": fD["ENUMS"],
                            "CONTENT_CLASSES": fD["CONTENT_CLASSES"],
                            "SUB_CATEGORIES": [qtD["id"] for qtD in fD["SUB_CATEGORIES"]],
                        }
                    )
                    atId = (convertNameF(atName)).upper()
                    dD["ATTRIBUTE_INFO"][atId] = td
                    atIdIndexList.append(atId)
                    atNameIndexList.append(atName)
                    #
                    mI = self.__contentInfo.getMethodImplementation(catName, atName, methodCodes=["calculate_on_load"])
                    if mI:
                        dD["ATTRIBUTE_MAP"].update({(convertNameF(atName)).upper(): {"CATEGORY": None, "ATTRIBUTE": None, "METHOD_NAME": mI, "ARGUMENTS": None}})
                    else:
                        dD["ATTRIBUTE_MAP"].update({(convertNameF(atName)).upper(): {"CATEGORY": catName, "ATTRIBUTE": atName, "METHOD_NAME": None, "ARGUMENTS": None}})
                    iOrder += 1
            #
            for atName in sorted(atNameList):
                fD = aD[atName]
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD["CONTENT_CLASSES"]):
                    continue

                if not fD["IS_KEY"]:
                    appType = dtAppInfo.getAppTypeName(fD["TYPE_CODE"])
                    if not appType:
                        logger.error("Missing data type mapping for %s %s", catName, atName)
                    appWidth = dtAppInfo.getAppTypeDefaultWidth(fD["TYPE_CODE"])
                    instWidth = dtInstInfo.getMaxWidth(catName, atName)
                    revAppType, revAppWidth = dtAppInfo.updateCharType(fD["IS_KEY"], appType, instWidth, appWidth, bufferPercent=20.0)
                    if verbose and dataTyping in ["SQL", "ANY"]:
                        logger.debug(
                            "catName %s atName %s cifType %s appType %s appWidth %r instWidth %r --> revAppType %r revAppWidth %r ",
                            catName,
                            atName,
                            fD["TYPE_CODE"],
                            appType,
                            appWidth,
                            instWidth,
                            revAppType,
                            revAppWidth,
                        )

                    #
                    appPrecision = dtAppInfo.getAppTypeDefaultPrecision(fD["TYPE_CODE"])
                    td = OrderedDict(
                        {
                            "ORDER": iOrder,
                            "NULLABLE": not fD["IS_MANDATORY"],
                            "PRECISION": appPrecision,
                            "PRIMARY_KEY": fD["IS_KEY"],
                            "APP_TYPE": revAppType,
                            "WIDTH": revAppWidth,
                            "ITERABLE_DELIMITER": fD["ITERABLE_DELIMITER"],
                            "FILTER_TYPES": fD["FILTER_TYPES"],
                            "IS_CHAR_TYPE": fD["IS_CHAR_TYPE"],
                            "ENUMERATION": fD["ENUMS"],
                            "CONTENT_CLASSES": fD["CONTENT_CLASSES"],
                            "SUB_CATEGORIES": [qtD["id"] for qtD in fD["SUB_CATEGORIES"]],
                        }
                    )
                    atId = (convertNameF(atName)).upper()
                    dD["ATTRIBUTE_INFO"][atId] = td

                    mI = self.__contentInfo.getMethodImplementation(catName, atName, methodCodes=["calculate_on_load"])
                    if mI:
                        dD["ATTRIBUTE_MAP"].update({(convertNameF(atName)).upper(): {"CATEGORY": None, "ATTRIBUTE": None, "METHOD_NAME": mI, "ARGUMENTS": None}})
                    else:
                        dD["ATTRIBUTE_MAP"].update({(convertNameF(atName)).upper(): {"CATEGORY": catName, "ATTRIBUTE": atName, "METHOD_NAME": None, "ARGUMENTS": None}})
                    iOrder += 1
            #
            atIdDelete = convertNameF(blockAttributeName).upper() if blockAttributeName else None
            dD["SCHEMA_DELETE_ATTRIBUTE"] = atIdDelete

            dD["INDICES"] = {"p1": {"TYPE": "UNIQUE", "ATTRIBUTES": tuple(atIdIndexList)}}
            if len(atIdIndexList) > 1:
                dD["INDICES"]["s1"] = {"TYPE": "SEARCH", "ATTRIBUTES": tuple([atIdDelete])}
            #
            # JDW -  Need to review attribute names here -
            dD["MAP_MERGE_INDICES"] = {catName: {"ATTRIBUTES": tuple(atNameIndexList), "TYPE": "EQUI-JOIN"}}
            # ----
            tD = OrderedDict()
            logger.debug("Slice names %r", sliceNames)
            for sliceName in sorted(sliceNames):
                sL = self.__contentInfo.getSliceAttributes(sliceName, catName)
                logger.debug("Slice attributes %r", sL)
                if sL:
                    # Convert names to IDs --
                    tL = []
                    for slD in sL:
                        pD = OrderedDict(
                            {
                                "PARENT_CATEGORY": convertNameF(slD["PARENT_CATEGORY_NAME"]).upper(),
                                "PARENT_ATTRIBUTE": convertNameF(slD["PARENT_ATTRIBUTE_NAME"]).upper(),
                                "CHILD_ATTRIBUTE": convertNameF(slD["CHILD_ATTRIBUTE_NAME"]).upper(),
                            }
                        )
                        tL.append(pD)
                    tD[sliceName] = tL
            dD["SLICE_ATTRIBUTES"] = tD
            #
            # ---- slice cardinality
            #
            dD["SLICE_UNIT_CARDINALITY"] = OrderedDict()
            sliceCardD = self.__contentInfo.getSliceUnitCardinalityForDatabase()
            logger.debug("Slice card dict %r", sliceCardD.items())
            for sliceName, catL in sliceCardD.items():
                if catName in catL:
                    dD["SLICE_UNIT_CARDINALITY"][sliceName] = True
                else:
                    dD["SLICE_UNIT_CARDINALITY"][sliceName] = False
            #
            dD["SLICE_CATEGORY_EXTRAS"] = OrderedDict()
            sliceCatD = self.__contentInfo.getSliceCategoryExtrasForDatabase()
            logger.debug("Slice category extra dict %r", sliceCatD.items())
            for sliceName, catL in sliceCatD.items():
                if catName in catL:
                    dD["SLICE_CATEGORY_EXTRAS"][sliceName] = True
                else:
                    dD["SLICE_CATEGORY_EXTRAS"][sliceName] = False
            #
            scL = []
            for atId in dD["ATTRIBUTE_INFO"]:
                scL.extend(dD["ATTRIBUTE_INFO"][atId]["SUB_CATEGORIES"])
            dD["SCHEMA_SUB_CATEGORIES"] = sorted(set(scL))
            #
            # Make attributes dict consistent with map ...
            dD["ATTRIBUTES"].update({atId: convertNameF(tD["ATTRIBUTE"]) for atId, tD in dD["ATTRIBUTE_MAP"].items() if atId not in dD["ATTRIBUTES"]})

            #
            rD[sId] = dD
        #
        return rD

    def __convertSliceParentItemNames(self, databaseName, dataTyping):
        sliceD = {}
        try:
            convertNameF = self.__getConvertNameMethod(dataTyping)
            # [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'id'}
            spD = self.__contentInfo.getSliceParentItemsForDatabase()
            for ky in spD:
                rL = []
                for aL in spD[ky]:
                    dD = {"CATEGORY": convertNameF(aL["CATEGORY_NAME"]).upper(), "ATTRIBUTE": convertNameF(aL["ATTRIBUTE_NAME"]).upper()}
                    rL.append(dD)
                sliceD[ky] = rL
            #
            return sliceD
        except Exception as e:
            logger.exception("Failing for %s with %s", databaseName, str(e))

        return sliceD

    def __convertSliceParentFilterNames(self, databaseName, dataTyping):
        sliceD = {}
        try:
            convertNameF = self.__getConvertNameMethod(dataTyping)
            # [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'id'}
            spD = self.__contentInfo.getSliceParentFiltersForDatabase()
            for ky in spD:
                rL = []
                for aL in spD[ky]:
                    dD = {"CATEGORY": convertNameF(aL["CATEGORY_NAME"]).upper(), "ATTRIBUTE": convertNameF(aL["ATTRIBUTE_NAME"]).upper(), "VALUES": aL["VALUES"]}
                    rL.append(dD)
                sliceD[ky] = rL
            #
            return sliceD
        except Exception as e:
            logger.exception("Failing for %s with %s", databaseName, str(e))

        return sliceD

    def __convertNameDefault(self, name):
        """ Default schema name converter -
        """
        logger.info("Using default name conversion for %r", name)
        return name

    # -------------------------- ------------- ------------- ------------- ------------- ------------- -------------
    def __getAmmendedAttributeFeatures(self, collectionName, catName, documentDefHelper):
        #
        useDefaultBrief = False
        aD = self.__contentInfo.getAttributeFeatures(catName)
        for atName, fD in aD.items():
            ascL = documentDefHelper.getAttributeSearchContexts(collectionName, catName, atName)
            fD["SEARCH_CONTEXTS"] = ascL
            # logger.info("collectionName %r catName %r atName %r context %r", collectionName, catName, atName, ascL)
            fD["IS_NESTED"] = documentDefHelper.isCategoryNested(collectionName, catName)
            fD["SEARCH_PRIORITY"] = documentDefHelper.getAttributeTextSearchPriority(collectionName, catName, atName)
            # logger.info("collectionName %r catName %r atName %r priority %r", collectionName, catName, atName, fD["SEARCH_PRIORITY"])
            tS = documentDefHelper.getAttributeDescription(catName, atName, contextType="brief")
            if tS:
                fD["DESCRIPTION_ANNOTATED"].append({"text": tS, "context": "brief"})
            elif useDefaultBrief:
                # Use content == deposition or then dictionary for 'brief' by default -
                for tD in fD["DESCRIPTION_ANNOTATED"]:
                    if tD["context"] == "deposition":
                        fD["DESCRIPTION_ANNOTATED"].append({"text": tD["text"], "context": "brief"})
                    elif tD["context"] == "dictionary":
                        fD["DESCRIPTION_ANNOTATED"].append({"text": tD["text"], "context": "brief"})
            # ----
            fD["SEARCH_GROUP_AND_PRIORITY"] = [{"group_name": tup[0], "priority_order": tup[1]} for tup in documentDefHelper.getSearchGroup(catName, atName)]
            # ----
        #
        return aD
        #

    def __exportSearchContext(self, collectionName, catName, atName, atPropD):
        if "bsonType" in atPropD:
            #
            sC = None
            tt = atPropD["bsonType"]
            if "enum" in atPropD:
                sC = "exact-match"
            elif "date" in tt or tt in ["int", "integer", "float", "double"]:
                sC = "default-match"
            elif tt in ["string"]:
                sC = "full-text"
            else:
                logger.warning("UNKNOWN type for %s %s %s", catName, atName, tt)
                return False

            logger.info("--SCONTEXT %s %s - %s.%s", collectionName, sC, catName, atName)
        return True

    def __createJsonLikeSchema(
        self,
        databaseName,
        collectionName,
        dataTyping,
        contentDefHelper,
        documentDefHelper,
        includeContentClasses=None,
        jsonSpecDraft="4",
        enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums",
        removeSubCategoryPrefix=True,
    ):
        """Internal method to integrate dictionary and instance metadata into a common json/bson schema description data structure.

           Working only for practical schema style: rowwise_by_name_with_cardinality

        Args:
            databaseName (str): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            collectionName (str): Collection defined within a schema/content type
            dataTyping (str): Target data type convention for the schema (e.g. JSON, BSON, or a variant of these...)
            contentDefHelper (class instance): Class instance providing additional schema details
            documentDefHelper (class instance): Class instance providing additional document schema details
            includeContentClasses (list, optional): list of additional content classes to be included (e.g. GENERATED_CONTENT)
            jsonSpecDraft (str, optional): The target draft schema specification '4|6'
            enforceOpts (str, optional): options for semantics are included in the schema (e.g. "mandatoryKeys|mandatoryAttributes|bounds|enums")

        Returns:
            dict: representation of JSON/BSON schema -


        """
        exportSearchContext = False
        addRcsbExtensions = "rcsb" in enforceOpts
        addBlockAttribute = True
        mandatorySubcategoryAttributes = True
        addPrimaryKey = "addPrimaryKey" in enforceOpts
        suppressSingleton = not documentDefHelper.getRetainSingletonObjects(collectionName)
        suppressRelations = documentDefHelper.getSuppressedCategoryRelationships(collectionName)
        logger.debug("Collection %s suppress singleton %r", collectionName, suppressSingleton)
        subCategoryAggregates = documentDefHelper.getSubCategoryAggregates(collectionName)
        logger.debug("%s %s Sub_category aggregates %r", databaseName, collectionName, subCategoryAggregates)
        privDocKeyL = documentDefHelper.getPrivateDocumentAttributes(collectionName)

        dataTypingU = dataTyping.upper()
        typeKey = "bsonType" if dataTypingU == "BSON" else "type"
        convertNameF = self.__getConvertNameMethod(dataTypingU)
        #
        contentClasses = includeContentClasses if includeContentClasses else []
        logger.debug("Including additional category classes %r", contentClasses)
        exportConfig = dataTypingU == "JSON" and False
        logger.debug("QQQQ COLLECTION %s", collectionName)
        if exportConfig:
            logger.info("CONFIG  %s:", collectionName)
        #
        dtInstInfo = self.__dtP.getDataTypeInstanceApi(databaseName)
        dtAppInfo = self.__dtP.getDataTypeApplicationApi(dataTypingU)
        #
        #      Supplied by the contentDefHelper for the content type (SchemaIds)
        #
        includeList = contentDefHelper.getIncluded(databaseName) if contentDefHelper else []
        excludeList = contentDefHelper.getExcluded(databaseName) if contentDefHelper else []
        excludeAttributesD = contentDefHelper.getExcludedAttributes(databaseName) if contentDefHelper else {}
        #
        #      Supplied by the documentDefHelp for the collection (SchemaIds)
        #
        docIncludeList = documentDefHelper.getIncluded(collectionName)
        docExcludeList = documentDefHelper.getExcluded(collectionName)
        docExcludeAttributes = documentDefHelper.getDocumentExcludedAttributes(collectionName)
        # JDW combine the schema and document excluded attributes here ...
        excludeAttributesD.update(docExcludeAttributes)

        sliceFilter = documentDefHelper.getSliceFilter(collectionName)
        sliceCategories = self.__contentInfo.getSliceCategories(sliceFilter) if sliceFilter else []
        sliceCategoryExtrasD = self.__contentInfo.getSliceCategoryExtrasForDatabase() if sliceFilter else {}
        if sliceFilter in sliceCategoryExtrasD:
            sliceCategories.extend(sliceCategoryExtrasD[sliceFilter])
        sliceCardD = self.__contentInfo.getSliceUnitCardinalityForDatabase() if sliceFilter else {}
        #
        if addBlockAttribute:
            # Optional synthetic attribute added to each category with value linked to data block identifier (or other function)
            blockAttributeName = contentDefHelper.getBlockAttributeName(databaseName) if contentDefHelper else None
            blockAttributeCifType = contentDefHelper.getBlockAttributeCifType(databaseName) if contentDefHelper else None
            blockAttributeAppType = dtAppInfo.getAppTypeName(blockAttributeCifType)
            blockRefPathList = None
            if "addParentRefs" in enforceOpts:
                refD = contentDefHelper.getBlockAttributeRefParent(databaseName)
                if refD is not None:
                    blockRefPathList = [convertNameF(refD["CATEGORY_NAME"]), convertNameF(refD["ATTRIBUTE_NAME"])]

        #
        dictSchema = self.__contentInfo.getSchemaNames()
        #
        sNameD = {}
        schemaPropD = {}
        mandatoryCategoryL = []
        for catName, fullAtNameList in dictSchema.items():
            atNameList = [at for at in fullAtNameList if (catName, at) not in excludeAttributesD]
            #

            #
            cfD = self.__contentInfo.getCategoryFeatures(catName)
            # logger.debug("catName %s contentClasses %r cfD %r" % (catName, contentClasses, cfD))

            #
            #  Skip categories that are uniformly unpopulated --
            #
            if not dtInstInfo.exists(catName) and not self.__testContentClasses(contentClasses, cfD["CONTENT_CLASSES"]):
                logger.debug("Schema %r Skipping category %s content classes %r", databaseName, catName, cfD["CONTENT_CLASSES"])
                continue
            #
            # -> Create a schema id  for catName <-
            sName = convertNameF(catName)
            sNameD[sName] = catName
            schemaId = sName.upper()
            #
            #  These are the content type schema level filters -
            if excludeList and schemaId in excludeList:
                continue
            if includeList and schemaId not in includeList:
                continue
            #
            # These are collection level filters
            #
            if docExcludeList and schemaId in docExcludeList:
                continue
            if docIncludeList and schemaId not in docIncludeList:
                continue
            #
            #  If there is a slice filter on this collection, the skip categories not connected to the slice
            if sliceFilter and catName not in sliceCategories:
                continue
            #
            #        Done with category filtering/selections
            # -------- ---------- ------------ -------- ---------- ------------ -------- ---------- ------------
            #
            # aD = self.__contentInfo.getAttributeFeatures(catName)
            aD = self.__getAmmendedAttributeFeatures(collectionName, catName, documentDefHelper)
            #
            if cfD["IS_MANDATORY"]:
                mandatoryCategoryL.append(catName)
            #
            isUnitCard = True if ("UNIT_CARDINALITY" in cfD and cfD["UNIT_CARDINALITY"]) else False
            if sliceFilter and sliceFilter in sliceCardD:
                isUnitCard = catName in sliceCardD[sliceFilter]
            #
            pD = {typeKey: "object", "properties": {}, "additionalProperties": False}
            #

            if isUnitCard:
                catPropD = pD
            else:
                if cfD["IS_MANDATORY"]:
                    catPropD = {typeKey: "array", "items": pD, "minItems": 1, "uniqueItems": True}
                else:
                    # JDW Adjusted minItems=1
                    catPropD = {typeKey: "array", "items": pD, "minItems": 1, "uniqueItems": True}
                #
            if dataTypingU == "JSON" and addRcsbExtensions:
                self.__updateCategoryNestedContext(catPropD, collectionName, catName, documentDefHelper)
                #

            if addBlockAttribute and blockAttributeName:
                schemaAttributeName = convertNameF(blockAttributeName)
                pD.setdefault("required", []).append(schemaAttributeName)
                #
                if blockRefPathList:
                    atPropD = self.__getJsonRef(blockRefPathList)
                else:
                    # atPropD = {typeKey: blockAttributeAppType, 'maxWidth': blockAttributeWidth}
                    atPropD = {typeKey: blockAttributeAppType}

                if addPrimaryKey:
                    atPropD["_primary_key"] = True
                #
                pD["properties"][schemaAttributeName] = atPropD

            #  First, filter any subcategory aggregates from the available list of a category attributes
            #

            subCatPropD = {}
            if subCategoryAggregates:
                logger.debug("%s %s %s subcategories %r", databaseName, collectionName, catName, cfD["SUB_CATEGORIES"])
                for subCategory in subCategoryAggregates:
                    if subCategory not in cfD["SUB_CATEGORIES"]:
                        continue
                    logger.debug("%s %s %s processing subcategory %r", databaseName, collectionName, catName, subCategory)
                    reqL = []
                    scD = {typeKey: "object", "properties": {}, "additionalProperties": False}
                    scHasUnitCard = documentDefHelper.getSubCategoryAggregateUnitCardinality(collectionName, subCategory)
                    for atName in sorted(atNameList):
                        fD = aD[atName]
                        #
                        if subCategory not in [qtD["id"] for qtD in fD["SUB_CATEGORIES"]]:
                            continue
                        #
                        schemaAttributeName = convertNameF(atName)
                        if removeSubCategoryPrefix:
                            schemaAttributeName = schemaAttributeName.replace(subCategory + "_", "")
                        #
                        # isRequired = "mandatoryAttributes" in enforceOpts and fD["IS_MANDATORY"]
                        # JDW separate general support for mandatory attributes and support in subcategories.
                        isRequired = mandatorySubcategoryAttributes and fD["IS_MANDATORY"]
                        if isRequired:
                            reqL.append(schemaAttributeName)
                        #
                        atPropD = self.__getJsonAttributeProperties(fD, dataTypingU, dtAppInfo, dtInstInfo, jsonSpecDraft, enforceOpts, suppressRelations, addRcsbExtensions)
                        scD["properties"][schemaAttributeName] = atPropD
                        if exportSearchContext:
                            self.__exportSearchContext(collectionName, catName, atName, atPropD)

                    if scD["properties"]:
                        if reqL:
                            scD["required"] = reqL
                        if scHasUnitCard:
                            subCatPropD[subCategory] = scD
                        else:
                            subCatPropD[subCategory] = {typeKey: "array", "items": scD, "uniqueItems": False}
                            if dataTypingU == "JSON" and addRcsbExtensions:
                                if documentDefHelper.isSubCategoryNested(collectionName, catName, subCategory):
                                    tD = documentDefHelper.getSubCategoryNestedContext(collectionName, catName, subCategory)
                                    if "FIRST_CONTEXT_PATH" in tD and tD["FIRST_CONTEXT_PATH"]:
                                        subCatPropD[subCategory]["rcsb_nested_indexing"] = True
                                        ###
                                        if "CONTEXT_ATTRIBUTE_VALUES" in tD and tD["CONTEXT_ATTRIBUTE_VALUES"]:
                                            vvDL = []
                                            for cavD in tD["CONTEXT_ATTRIBUTE_VALUES"]:
                                                vvD = {"context_value": cavD["CONTEXT_VALUE"], "search_paths": cavD["SEARCH_PATHS"]}
                                                vvDL.append(vvD)
                                            subCatPropD[subCategory]["rcsb_nested_indexing_context"] = [
                                                {"category_name": tD["CONTEXT_NAME"], "category_path": tD["FIRST_CONTEXT_PATH"], "context_attributes": vvDL}
                                            ]
                                        else:
                                            subCatPropD[subCategory]["rcsb_nested_indexing_context"] = [
                                                {"category_name": tD["CONTEXT_NAME"], "category_path": tD["FIRST_CONTEXT_PATH"]}
                                            ]
                                    else:
                                        subCatPropD[subCategory]["rcsb_nested_indexing"] = True
            #
            if subCatPropD:
                logger.debug("%s %s %s processing subcategory properties %r", databaseName, collectionName, catName, subCatPropD.items())
            #
            if exportConfig:
                logger.info("CONFIG - CATEGORY_NAME: %s", catName)
                logger.info("CONFIG   ATTRIBUTE_NAME_LIST:")
            for atName in sorted(atNameList):
                fD = aD[atName]
                # Exclude primary data attributes with no instance coverage except if in a protected content class
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD["CONTENT_CLASSES"]):
                    continue
                if subCategoryAggregates and self.__subCategoryTest(subCategoryAggregates, [d["id"] for d in fD["SUB_CATEGORIES"]]):
                    continue
                #
                if exportConfig and "_id" in atName:
                    logger.info("CONFIG   - %s", atName)
                #
                schemaAttributeName = convertNameF(atName)
                isRequired = ("mandatoryKeys" in enforceOpts and fD["IS_KEY"]) or ("mandatoryAttributes" in enforceOpts and fD["IS_MANDATORY"])
                # subject to exclusion
                if isRequired and (catName, atName) not in excludeAttributesD:
                    pD.setdefault("required", []).append(schemaAttributeName)
                #
                atPropD = self.__getJsonAttributeProperties(fD, dataTypingU, dtAppInfo, dtInstInfo, jsonSpecDraft, enforceOpts, suppressRelations, addRcsbExtensions)

                delimiter = fD["ITERABLE_DELIMITER"]
                if delimiter:
                    pD["properties"][schemaAttributeName] = {typeKey: "array", "items": atPropD, "uniqueItems": False}
                else:
                    pD["properties"][schemaAttributeName] = atPropD

                if exportSearchContext:
                    self.__exportSearchContext(collectionName, catName, schemaAttributeName, atPropD)

            if subCatPropD:
                pD["properties"].update(copy.copy(subCatPropD))
            # pD['required'].extend(list(subCatPropD.keys()))
            #
            if "required" in catPropD and not catPropD["required"]:
                logger.debug("Category %s cfD %r", catName, cfD.items())
                del catPropD["required"]
            #
            if pD["properties"]:
                schemaPropD[sName] = copy.deepcopy(catPropD)
        #
        # Add any private keys to the object schema - Fetch the metadata for the private keys
        #
        privKeyD = {}
        privMandatoryD = {}
        if privDocKeyL:
            for pdk in privDocKeyL:
                catNameK = convertNameF(pdk["CATEGORY_NAME"])
                aD = self.__contentInfo.getAttributeFeatures(catNameK)
                atNameK = convertNameF(pdk["ATTRIBUTE_NAME"])
                fD = aD[atNameK]
                atPropD = self.__getJsonAttributeProperties(fD, dataTypingU, dtAppInfo, dtInstInfo, jsonSpecDraft, enforceOpts, suppressRelations, addRcsbExtensions)
                privKeyD[pdk["PRIVATE_DOCUMENT_NAME"]] = atPropD
                privMandatoryD[pdk["PRIVATE_DOCUMENT_NAME"]] = pdk["MANDATORY"]
        #
        # Suppress the category name for schemas with a single category -
        #
        if suppressSingleton and len(schemaPropD) == 1:
            logger.debug("%s %s suppressing category in singleton schema", databaseName, collectionName)
            sName = list(schemaPropD.keys())[0]
            catName = sNameD[sName]
            logger.debug("%s singleton state sName %r catName %r", collectionName, sName, catName)
            # rD = copy.deepcopy(catPropD)
            for k, v in privKeyD.items():
                pD["properties"][k] = v
                # pD['required'] = k
                if privMandatoryD[k]:
                    pD.setdefault("required", []).append(k)
            rD = copy.deepcopy(pD)
            # if "additionalProperties" in rD:
            #    rD["additionalProperties"] = True
            if dataTypingU == "JSON" and addRcsbExtensions:
                self.__updateCategoryNestedContext(rD, collectionName, catName, documentDefHelper)
            logger.debug("%s singleton state rD %r", collectionName, rD)
        else:
            for k, v in privKeyD.items():
                schemaPropD[k] = v
                if privMandatoryD[k]:
                    mandatoryCategoryL.append(k)
            #
            rD = {typeKey: "object", "properties": schemaPropD, "additionalProperties": False}
            if mandatoryCategoryL:
                rD["required"] = mandatoryCategoryL

        #
        if dataTypingU == "BSON":
            rD["properties"]["_id"] = {"bsonType": "objectId"}
            logger.debug("Adding mongo key %r", rD["properties"]["_id"])
        #
        if dataTypingU == "JSON":
            sdType = dataTyping.lower()
            sLevel = "full" if "bounds" in enforceOpts else "min"
            fn = "%s-schema-%s-%s.json" % (sdType, sLevel, collectionName)
            collectionVersion = documentDefHelper.getCollectionVersion(databaseName, collectionName)
            jsonSchemaUrl = "http://json-schema.org/draft-0%s/schema#" % jsonSpecDraft if jsonSpecDraft in ["3", "4", "6", "7"] else "http://json-schema.org/schema#"
            schemaRepo = "https://github.com/rcsb/py-rcsb.db/tree/master/rcsb.db/data/json-schema/"
            desc1 = "RCSB Exchange Database JSON schema derived from the %s content type schema. " % databaseName
            desc2 = "This schema supports collection %s version %s. " % (collectionName, collectionVersion)
            desc3 = "This schema is hosted in repository %s%s and follows JSON schema specification version %s" % (schemaRepo, fn, jsonSpecDraft)
            rD.update(
                {
                    "$id": "%s%s" % (schemaRepo, fn),
                    "$schema": jsonSchemaUrl,
                    "title": "schema: %s collection: %s version: %s" % (databaseName, collectionName, collectionVersion),
                    "description": desc1 + desc2 + desc3,
                    "$comment": "schema_version: %s" % collectionVersion,
                }
            )

        return rD

    def __updateCategoryNestedContext(self, rD, collectionName, catName, documentDefHelper):
        """Insert nested context details into a category schema object

        Args:
            rD (dict): current category schema object
            collectionName (string): collection name
            catName (str): data category Name
            documentDefHelper (obj): instance of DocumentDefHelper()

        Returns:
            dict : updated category schema object

        Output schema example:

        "rcsb_nested_indexing_context": [
                {
                "category_name": "annotation_type",
                "category_path": "rcsb_polymer_entity_annotation.type"
                "context_attributes": [
                    {
                        "context_value": 'vxxxxx',
                        "search_paths": [ 'p1', 'p2', 'p3']
                    }
                ]
                }
            ]

                 CONTEXT_ATTRIBUTE_VALUES:
                    - CONTEXT_VALUE: a_value
                      SEARCH_PATHS:
                      - a.b
                      - c.d
        """
        if documentDefHelper.isCategoryNested(collectionName, catName):
            tD = documentDefHelper.getCategoryNestedContext(collectionName, catName)
            logger.debug("%s Nested context dict %r", collectionName, tD)
            if "FIRST_CONTEXT_PATH" in tD and tD["FIRST_CONTEXT_PATH"]:
                rD["rcsb_nested_indexing"] = True
                if "CONTEXT_ATTRIBUTE_VALUES" in tD and tD["CONTEXT_ATTRIBUTE_VALUES"]:
                    vDL = []
                    for cavD in tD["CONTEXT_ATTRIBUTE_VALUES"]:
                        vD = {"context_value": cavD["CONTEXT_VALUE"], "search_paths": cavD["SEARCH_PATHS"]}
                        vDL.append(vD)
                    rD["rcsb_nested_indexing_context"] = [{"category_name": tD["CONTEXT_NAME"], "category_path": tD["FIRST_CONTEXT_PATH"], "context_attributes": vDL}]
                else:
                    rD["rcsb_nested_indexing_context"] = [{"category_name": tD["CONTEXT_NAME"], "category_path": tD["FIRST_CONTEXT_PATH"]}]
                #
            else:
                rD["rcsb_nested_indexing"] = True
        return rD

    def __getJsonRef(self, pathList, itemSubCategoryList=None):
        """ Add parent reference and optional group membership and labeling.
        """
        refD = {}
        try:
            refD = {"$ref": "#" + "/".join(pathList)}
            if itemSubCategoryList:
                logger.debug("Subcategory membership %r %r", pathList, itemSubCategoryList)
                refD["_attribute_groups"] = itemSubCategoryList
        except Exception as e:
            logger.excepion("Failing with pathList %r %s" % (pathList, str(e)))
        return refD

    def __testSuppressRelation(self, fD, suppressRelations):
        ret = False
        childCatName = fD["CATEGORY_NAME"]
        parentCatName = fD["PARENT"]["CATEGORY"]
        for srD in suppressRelations:
            if "CHILD_CATEGORY_NAME" in srD and "PARENT_CATEGORY_NAME" in srD:
                ret = (childCatName == srD["CHILD_CATEGORY_NAME"] and parentCatName == srD["PARENT_CATEGORY_NAME"]) or ret
            elif "PARENT_CATEGORY_NAME" in srD:
                ret = (parentCatName == srD["PARENT_CATEGORY_NAME"]) or ret
            elif "CHILD_CATEGORY_NAME" in srD:
                ret = (childCatName == srD["CHILD_CATEGORY_NAME"]) or ret
        if ret:
            logger.debug("Suppressing relationships for catName %r parent %r", childCatName, parentCatName)

        return ret

    def __getJsonAttributeProperties(self, fD, dataTypingU, dtAppInfo, dtInstInfo, jsonSpecDraft, enforceOpts, suppressRelations, addRcsbExtensions):
        #
        atPropD = {}
        addPrimaryKey = "addPrimaryKey" in enforceOpts
        addParentRefs = "addParentRefs" in enforceOpts
        try:
            catName = fD["CATEGORY_NAME"]
            atName = fD["ATTRIBUTE_NAME"]
            precMin = dtInstInfo.getMinPrecision(catName, atName)
            precMax = dtInstInfo.getMaxPrecision(catName, atName)
            # Adding a parent reference -
            if addParentRefs and fD["PARENT"] is not None and not self.__testSuppressRelation(fD, suppressRelations):
                convertNameF = self.__getConvertNameMethod(dataTypingU)
                pCatName = convertNameF(fD["PARENT"]["CATEGORY"])
                pAtName = convertNameF(fD["PARENT"]["ATTRIBUTE"])
                # logger.info("Using parent ref %r %r " % (pCatName, pAtName))
                # atPropD = {'$ref': '#/%s/%s' % (pCatName, pAtName)}
                atPropD = self.__getJsonRef([pCatName, pAtName], fD["SUB_CATEGORIES"])
                if addPrimaryKey and fD["IS_KEY"]:
                    atPropD["_primary_key"] = True
                return atPropD
            #
            # - assign data type attributes
            typeKey = "bsonType" if dataTypingU == "BSON" else "type"
            appType = dtAppInfo.getAppTypeName(fD["TYPE_CODE"])
            #
            #
            if appType in ["string"]:
                # atPropD = {typeKey: appType, 'maxWidth': instWidth}
                atPropD = {typeKey: appType}
            elif appType in ["date", "datetime"] and dataTypingU == "JSON":
                fmt = "date" if appType == "date" else "date-time"
                atPropD = {typeKey: "string", "format": fmt}
            elif appType in ["date", "datetime"] and dataTypingU == "BSON":
                atPropD = {typeKey: "date"}
            elif appType in ["number", "integer", "int", "double"]:
                atPropD = {typeKey: appType}
                #

                #
                if "bounds" in enforceOpts:
                    if jsonSpecDraft in ["3", "4"]:
                        if "MIN_VALUE" in fD:
                            atPropD["minimum"] = fD["MIN_VALUE"]
                        elif "MIN_VALUE_EXCLUSIVE" in fD:
                            atPropD["minimum"] = fD["MIN_VALUE_EXCLUSIVE"]
                            atPropD["exclusiveMinimum"] = True
                        if "MAX_VALUE" in fD:
                            atPropD["maximum"] = fD["MAX_VALUE"]
                        elif "MAX_VALUE_EXCLUSIVE" in fD:
                            atPropD["maximum"] = fD["MAX_VALUE_EXCLUSIVE"]
                            atPropD["exclusiveMaximum"] = True
                    elif jsonSpecDraft in ["6", "7"]:
                        if "MIN_VALUE" in fD:
                            atPropD["minimum"] = fD["MIN_VALUE"]
                        elif "MIN_VALUE_EXCLUSIVE" in fD:
                            atPropD["exclusiveMinimum"] = fD["MIN_VALUE_EXCLUSIVE"]
                        if "MAX_VALUE" in fD:
                            atPropD["maximum"] = fD["MAX_VALUE"]
                        elif "MAX_VALUE_EXCLUSIVE" in fD:
                            atPropD["exclusiveMaximum"] = fD["MAX_VALUE_EXCLUSIVE"]
            else:
                atPropD = {typeKey: appType}

            if "enums" in enforceOpts and fD["ENUMS"]:
                atPropD["enum"] = sorted(fD["ENUMS"])
            #
            if dataTypingU not in ["BSON"]:
                try:
                    if fD["EXAMPLES"]:
                        atPropD["examples"] = [str(t1).strip() for t1, _ in fD["EXAMPLES"]]
                except Exception as e:
                    logger.exception("Failing for %r with %s", fD["EXAMPLES"], str(e))
                if fD["DESCRIPTION"]:
                    atPropD["description"] = fD["DESCRIPTION"]
                #
            if addPrimaryKey and fD["IS_KEY"]:
                atPropD["_primary_key"] = True
            if addParentRefs and fD["SUB_CATEGORIES"] and fD["SUB_CATEGORIES"]:
                atPropD["_attribute_groups"] = fD["SUB_CATEGORIES"]
            #
            if addRcsbExtensions and dataTypingU not in ["BSON"]:
                #
                if appType in ["double"] and precMin and precMax:
                    # need median here -
                    # atPropD["rcsb_precision_digits"] = precMin
                    pass

                if fD["SEARCH_CONTEXTS"]:
                    atPropD["rcsb_search_context"] = fD["SEARCH_CONTEXTS"]
                if "SEARCH_PRIORITY" in fD and fD["SEARCH_PRIORITY"]:
                    atPropD["rcsb_full_text_priority"] = int(fD["SEARCH_PRIORITY"])
                if fD["UNITS"]:
                    atPropD["rcsb_units"] = fD["UNITS"]
                #
                if fD["ENUMS_ANNOTATED"]:
                    atPropD["rcsb_enum_annotated"] = fD["ENUMS_ANNOTATED"]
                if fD["DESCRIPTION_ANNOTATED"]:
                    atPropD["rcsb_description"] = fD["DESCRIPTION_ANNOTATED"]
                if fD["SEARCH_GROUP_AND_PRIORITY"]:
                    atPropD["rcsb_search_group"] = fD["SEARCH_GROUP_AND_PRIORITY"]
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return atPropD

    def __subCategoryTest(self, filterList, atSubCategoryList):
        """ Return true if any element of filter list in atSubCategoryList
        """
        if not filterList or not atSubCategoryList:
            return False
        for subCat in filterList:
            if subCat in atSubCategoryList:
                return True
        return False



##
# File:    DictMethodChemRefHelper.py
# Author:  J. Westbrook
# Date:    16-Jul-2019
# Version: 0.001 Initial version
#
##
"""
Helper class implements external method references supporting chemical
reference data definitions in the RCSB dictionary extension.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

# from collections import Counter, OrderedDict

from mmcif.api.DataCategory import DataCategory

logger = logging.getLogger(__name__)


class DictMethodChemRefHelper(object):
    """ Helper class implements external method references supporting chemical
        reference data definitions in the RCSB dictionary extension.
    """

    def __init__(self, **kwargs):
        """
        Args:
            resourceProvider: (obj) instance of DictMethodResourceProvider()

        """
        #
        self._raiseExceptions = kwargs.get("raiseExceptions", False)
        #
        rP = kwargs.get("resourceProvider")
        self.__dApi = rP.getResource("Dictionary API instance (pdbx_core)") if rP else None
        logger.debug("Dictionary method helper init")

    def echo(self, msg):
        logger.info(msg)

    def addChemCompRelated(self, dataContainer, catName, **kwargs):
        """ Add category rcsb_chem_comp_related.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example,

             loop_
             _rcsb_chem_comp_related.comp_id
             _rcsb_chem_comp_related.ordinal
             _rcsb_chem_comp_related.resource_name
             _rcsb_chem_comp_related.resource_accession_code
             _rcsb_chem_comp_related.related_mapping_method
             ATP 1 DrugBank DB00171 'assigned by resource'
        """
        try:
            logger.debug("Starting with  %r %r", dataContainer.getName(), catName)
            # Exit if source categories are missing
            if not (dataContainer.exists("chem_comp_atom") and dataContainer.exists("chem_comp_bond")):
                return False
            rP = kwargs.get("resourceProvider")
            # ------- new
            ccId = self.__getChemCompId(dataContainer)
            dbId, atcIdL, mappingType, dbVersion = self.__getDrugBankMapping(dataContainer, rP)
            logger.debug("Using DrugBank version %r", dbVersion)
            #  ------------ ----------------------- ----------------------- ----------------------- -----------
            if dbId:
                #
                if dataContainer.exists("rcsb_chem_comp_container_identifiers"):
                    tObj = dataContainer.getObj("rcsb_chem_comp_container_identifiers")
                    if not tObj.hasAttribute("drugbank_id"):
                        tObj.appendAttribute("drugbank_id")
                    tObj.setValue(dbId, "drugbank_id", 0)
                    if atcIdL:
                        if not tObj.hasAttribute("atc_codes"):
                            tObj.appendAttribute("atc_codes")
                        tObj.setValue(",".join(atcIdL), "atc_codes", 0)
                #
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
                wObj = dataContainer.getObj(catName)
                rL = wObj.selectIndices("DrugBank", "resource_name")
                ok = False
                if rL:
                    ok = wObj.removeRows(rL)
                    if not ok:
                        logger.debug("Error removing rows in %r %r", catName, rL)
                # ---
                iRow = wObj.getRowCount()
                wObj.setValue(ccId, "comp_id", iRow)
                wObj.setValue(iRow + 1, "ordinal", iRow)
                wObj.setValue("DrugBank", "resource_name", iRow)
                wObj.setValue(dbId, "resource_accession_code", iRow)
                wObj.setValue(mappingType, "related_mapping_method", iRow)
                #
            #  ------------ ----------------------- ----------------------- ----------------------- -----------
            ccmProvider = rP.getResource("ChemCompModelProvider instance") if rP else None
            csdMapD = ccmProvider.getMapping()
            #
            if csdMapD and ccId in csdMapD:
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
                wObj = dataContainer.getObj(catName)
                logger.debug("Using CSD model mapping length %d", len(csdMapD))
                dbId = csdMapD[ccId][0]["db_code"]
                rL = wObj.selectIndices("CCDC/CSD", "resource_name")
                if rL:
                    ok = wObj.removeRows(rL)
                    if not ok:
                        logger.debug("Error removing rows in %r %r", catName, rL)
                iRow = wObj.getRowCount()
                wObj.setValue(ccId, "comp_id", iRow)
                wObj.setValue(iRow + 1, "ordinal", iRow)
                wObj.setValue("CCDC/CSD", "resource_name", iRow)
                wObj.setValue(dbId, "resource_accession_code", iRow)
                wObj.setValue("assigned by PDB", "related_mapping_method", iRow)
            #
            residProvider = rP.getResource("ResidProvider instance") if rP else None
            residMapD = residProvider.getMapping()
            #
            if residMapD and ccId in residMapD:
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
                wObj = dataContainer.getObj(catName)
                rL = wObj.selectIndices("RESID", "resource_name")
                if rL:
                    ok = wObj.removeRows(rL)
                    if not ok:
                        logger.debug("Error removing rows in %r %r", catName, rL)
                logger.debug("Using RESID model mapping length %d", len(residMapD))
                for rD in residMapD[ccId]:
                    dbId = rD["residCode"]
                    iRow = wObj.getRowCount()
                    wObj.setValue(ccId, "comp_id", iRow)
                    wObj.setValue(iRow + 1, "ordinal", iRow)
                    wObj.setValue("RESID", "resource_name", iRow)
                    wObj.setValue(dbId, "resource_accession_code", iRow)
                    wObj.setValue("matching by RESID resource", "related_mapping_method", iRow)
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def __getChemCompId(self, dataContainer):
        if not dataContainer.exists("chem_comp"):
            return None
        ccObj = dataContainer.getObj("chem_comp")
        if not ccObj.hasAttribute("pdbx_release_status"):
            return None
        return ccObj.getValueOrDefault("id", 0, None)

    def __getDrugBankMapping(self, dataContainer, resourceProvider):
        """ Return the DrugBank mapping for the chemical definition in the input dataContainer.

        Args:
            dataContainer (obj): instance of a DataContainer() object
            resourceProvider (obj): instance of a ResourceProvider() object

        Returns:
            mType, DrugBankId, actL (str,str, list): mapping type and DrugBank accession code, list of ATC assignments
        """
        try:
            dbId = None
            atcL = []
            mappingType = None

            dbProvider = resourceProvider.getResource("DrugBankProvider instance") if resourceProvider else None
            dbD = dbProvider.getMapping()
            dbVersion = dbProvider.getVersion()
            if dbD:
                ccId = self.__getChemCompId(dataContainer)
                #
                dbMapD = dbD["id_map"]
                inKeyD = dbD["inchikey_map"]
                atcD = dbD["db_atc_map"]
                logger.debug("atcD length is %d", len(atcD))
                logger.debug("inKeyD length is %d", len(inKeyD))
                #
                if dataContainer.exists("rcsb_chem_comp_descriptor"):
                    ccIObj = dataContainer.getObj("rcsb_chem_comp_descriptor")

                    if ccIObj.hasAttribute("InChIKey"):
                        inky = ccIObj.getValue("InChIKey", 0)
                        logger.debug("inKeyD length is %d testing %r", len(inKeyD), inky)
                        if inky in inKeyD:
                            logger.debug("Matching inchikey for %s", ccId)
                            dbId = inKeyD[inky][0]["drugbank_id"]
                            mappingType = "matching InChIKey in DrugBank"
                #

                if not dbId and dbMapD and dataContainer.getName() in dbMapD:
                    dbId = dbMapD[ccId]["drugbank_id"]
                    mappingType = "assigned by DrugBank resource"
                    logger.debug("Matching db assignment for %s", ccId)
                if atcD and dbId in atcD:
                    atcL = atcD[dbId]

        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return dbId, atcL, mappingType, dbVersion

    def addChemCompAnnotation(self, dataContainer, catName, **kwargs):
        """Generate the rcsb_chem_annotation category -

        Args:
            dataContainer ([type]): [description]
            catName ([type]): [description]

        Returns:
            [type]: [description]

                loop_
                _rcsb_chem_comp_annotation.ordinal
                _rcsb_chem_comp_annotation.entry_id
                _rcsb_chem_comp_annotation.entity_id
                #
                _rcsb_chem_comp_annotation.annotation_id
                _rcsb_chem_comp_annotation.type
                _rcsb_chem_comp_annotation.name
                _rcsb_chem_comp_annotation.description
                #
                _rcsb_chem_comp_annotation.annotation_lineage_id
                _rcsb_chem_comp_annotation.annotation_lineage_name
                _rcsb_chem_comp_annotation.annotation_lineage_depth
                #
                _rcsb_chem_comp_annotation.provenance_source
                _rcsb_chem_comp_annotation.assignment_version
                # ...
        """
        try:
            # Exit if these source categories are missing
            if not (dataContainer.exists("chem_comp_atom") and dataContainer.exists("chem_comp_bond")):
                return False
            #
            logger.debug("Starting with  %r %r", dataContainer.getName(), catName)
            rP = kwargs.get("resourceProvider")
            #
            # ----
            ccId = self.__getChemCompId(dataContainer)
            dbId, atcIdL, mappingType, dbVersion = self.__getDrugBankMapping(dataContainer, rP)
            atcP = rP.getResource("AtcProvider instance") if rP else None
            if atcIdL and atcP:
                #
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
                # -----
                wObj = dataContainer.getObj(catName)
                #
                for atcId in atcIdL:
                    iRow = wObj.getRowCount()
                    wObj.setValue(ccId, "comp_id", iRow)
                    wObj.setValue(iRow + 1, "ordinal", iRow)
                    wObj.setValue("ATC", "type", iRow)
                    wObj.setValue(atcId, "annotation_id", iRow)
                    wObj.setValue(atcP.getAtcName(atcId), "name", iRow)
                    #
                    wObj.setValue("ATC " + mappingType, "description", iRow)
                    # ---
                    wObj.setValue(";".join(atcP.getNameLineage(atcId)), "annotation_lineage_name", iRow)
                    idLinL = atcP.getIdLineage(atcId)
                    wObj.setValue(";".join(idLinL), "annotation_lineage_id", iRow)
                    wObj.setValue(";".join([str(jj) for jj in range(0, len(idLinL) + 1)]), "annotation_lineage_depth", iRow)
                    #
                    wObj.setValue("DrugBank", "provenance_source", iRow)
                    wObj.setValue(dbVersion, "assignment_version", iRow)
                    logger.debug("dbId %r atcId %r lineage %r", dbId, atcId, idLinL)
            # -----
            rsProvider = rP.getResource("ResidProvider instance") if rP else None
            residD = rsProvider.getMapping()
            #
            if residD and (ccId in residD):
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
                wObj = dataContainer.getObj(catName)
                # -----
                residVersion = rsProvider.getVersion()
                jj = 1
                for rD in residD[ccId]:
                    if "modRes" not in rD:
                        continue
                    for modRes in rD["modRes"]:
                        iRow = wObj.getRowCount()
                        wObj.setValue(ccId, "comp_id", iRow)
                        wObj.setValue(iRow + 1, "ordinal", iRow)
                        wObj.setValue("Modification Type", "type", iRow)
                        wObj.setValue("modres_%d" % jj, "annotation_id", iRow)
                        wObj.setValue(modRes, "name", iRow)
                        wObj.setValue("RESID", "provenance_source", iRow)
                        wObj.setValue(residVersion, "assignment_version", iRow)
                        jj += 1
                #
                jj = 1
                for rD in residD[ccId]:
                    if "genEnzymes" not in rD:
                        continue
                    for genEnzyme in rD["genEnzymes"]:
                        iRow = wObj.getRowCount()
                        wObj.setValue(ccId, "comp_id", iRow)
                        wObj.setValue(iRow + 1, "ordinal", iRow)
                        wObj.setValue("Generating Enzyme", "type", iRow)
                        wObj.setValue("enzyme_%d" % jj, "annotation_id", iRow)
                        wObj.setValue(genEnzyme, "name", iRow)
                        wObj.setValue("RESID", "provenance_source", iRow)
                        wObj.setValue(residVersion, "assignment_version", iRow)
                        jj += 1
                #
                psimodP = rP.getResource("PsiModProvider instance") if rP else None
                if psimodP:
                    jj = 1
                    for rD in residD[ccId]:
                        if "ontRefs" not in rD:
                            continue
                        for ontId in rD["ontRefs"]:
                            if ontId[:3] != "MOD":
                                continue
                            iRow = wObj.getRowCount()
                            wObj.setValue(ccId, "comp_id", iRow)
                            wObj.setValue(iRow + 1, "ordinal", iRow)
                            wObj.setValue("PSI-MOD", "type", iRow)
                            wObj.setValue(ontId, "annotation_id", iRow)
                            wObj.setValue(psimodP.getName(ontId), "name", iRow)
                            wObj.setValue("RESID", "provenance_source", iRow)
                            wObj.setValue(residVersion, "assignment_version", iRow)
                            #
                            linL = psimodP.getLineage(ontId)
                            wObj.setValue(";".join([tup[0] for tup in linL]), "annotation_lineage_id", iRow)
                            wObj.setValue(";".join([tup[1] for tup in linL]), "annotation_lineage_name", iRow)
                            wObj.setValue(";".join([str(tup[2]) for tup in linL]), "annotation_lineage_depth", iRow)
                    #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def addChemCompTargets(self, dataContainer, catName, **kwargs):
        """Add category rcsb_chem_comp_target using DrugBank annotations.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        Example:
             loop_
             _rcsb_chem_comp_target.comp_id
             _rcsb_chem_comp_target.ordinal
             _rcsb_chem_comp_target.name
             _rcsb_chem_comp_target.interaction_type
             _rcsb_chem_comp_target.target_actions
             _rcsb_chem_comp_target.organism_common_name
             _rcsb_chem_comp_target.reference_database_name
             _rcsb_chem_comp_target.reference_database_accession_code
             _rcsb_chem_comp_target.provenance_code
             ATP 1 "O-phosphoseryl-tRNA(Sec) selenium transferase" target cofactor Human UniProt Q9HD40 DrugBank

        DrugBank target info:
        {
            "type": "target",
            "name": "Alanine--glyoxylate aminotransferase 2, mitochondrial",
            "organism": "Human",
            "actions": [
               "cofactor"
            ],
            "known_action": "unknown",
            "uniprot_ids": "Q9BYV1"
         },

        """
        try:
            logger.debug("Starting with  %r %r", dataContainer.getName(), catName)
            # Exit if source categories are missing
            if not (dataContainer.exists("chem_comp_atom") and dataContainer.exists("chem_comp_bond")):
                return False

            #
            rP = kwargs.get("resourceProvider")
            dbProvider = rP.getResource("DrugBankProvider instance") if rP else None
            dbD = dbProvider.getMapping()
            if not dbD:
                return False

            dbMapD = dbD["id_map"] if "id_map" in dbD else None
            #
            ccId = dataContainer.getName()
            if dbMapD and ccId in dbMapD and "target_interactions" in dbMapD[ccId]:
                #
                # Create the new target category
                if not dataContainer.exists(catName):
                    dataContainer.append(
                        DataCategory(
                            catName,
                            attributeNameList=[
                                "comp_id",
                                "ordinal",
                                "name",
                                "interaction_type",
                                "target_actions",
                                "organism_common_name",
                                "reference_database_name",
                                "reference_database_accession_code",
                                "provenance_code",
                            ],
                        )
                    )
                wObj = dataContainer.getObj(catName)
                logger.debug("Using DrugBank mapping length %d", len(dbMapD))
                rL = wObj.selectIndices("DrugBank", "provenance_code")
                if rL:
                    ok = wObj.removeRows(rL)
                    if not ok:
                        logger.debug("Error removing rows in %r %r", catName, rL)
                #
                iRow = wObj.getRowCount()
                iRow = wObj.getRowCount()
                for tD in dbMapD[ccId]["target_interactions"]:
                    wObj.setValue(ccId, "comp_id", iRow)
                    wObj.setValue(iRow + 1, "ordinal", iRow)
                    wObj.setValue(tD["name"], "name", iRow)
                    wObj.setValue(tD["type"], "interaction_type", iRow)
                    if "actions" in tD and tD["actions"]:
                        wObj.setValue(";".join(tD["actions"]), "target_actions", iRow)
                    if "organism" in tD:
                        wObj.setValue(tD["organism"], "organism_common_name", iRow)
                    if "uniprot_ids" in tD:
                        wObj.setValue("UniProt", "reference_database_name", iRow)
                        wObj.setValue(tD["uniprot_ids"], "reference_database_accession_code", iRow)
                    wObj.setValue("DrugBank", "provenance_code", iRow)
                    iRow += 1

            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def __getAuditDates(self, dataContainer, catName):
        createDate = None
        releaseDate = None
        reviseDate = None
        try:
            if dataContainer.exists(catName):
                cObj = dataContainer.getObj(catName)
                for iRow in range(cObj.getRowCount()):
                    aType = cObj.getValueOrDefault("action_type", iRow, defaultValue=None)
                    dateVal = cObj.getValueOrDefault("date", iRow, defaultValue=None)
                    if aType in ["Create component"]:
                        createDate = dateVal
                    elif aType in ["Initial release"]:
                        releaseDate = dateVal
                reviseDate = cObj.getValueOrDefault("date", cObj.getRowCount() - 1, defaultValue=None)
        except Exception as e:
            logger.exception("Faling with %s", str(e))
        return createDate, releaseDate, reviseDate

    def addChemCompInfo(self, dataContainer, catName, **kwargs):
        """ Add category rcsb_chem_comp_info and rcsb_chem_comp_container_identifiers.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example,
             _rcsb_chem_comp_info.comp_id                 BNZ
             _rcsb_chem_comp_info.atom_count              12
             _rcsb_chem_comp_info.atom_count_chiral        0
             _rcsb_chem_comp_info.bond_count              12
             _rcsb_chem_comp_info.bond_count_aromatic      6
             _rcsb_chem_comp_info.atom_count_heavy         6
        """
        try:
            logger.debug("Starting with  %r %r %r", dataContainer.getName(), catName, kwargs)
            # Exit if source categories are missing
            if not dataContainer.exists("chem_comp"):
                return False
            ccObj = dataContainer.getObj("chem_comp")
            if not ccObj.hasAttribute("pdbx_release_status"):
                return False
            ccId = ccObj.getValue("id", 0)
            ccReleaseStatus = ccObj.getValue("pdbx_release_status", 0)
            subComponentIds = ccObj.getValueOrDefault("pdbx_subcomponent_list", 0, defaultValue=None)
            #
            #
            prdId = prdReleaseStatus = representAs = None
            if dataContainer.exists("pdbx_reference_molecule"):
                prdObj = dataContainer.getObj("pdbx_reference_molecule")
                prdId = prdObj.getValueOrDefault("prd_id", 0, defaultValue=None)
                prdReleaseStatus = prdObj.getValueOrDefault("release_status", 0, defaultValue=None)
                representAs = prdObj.getValueOrDefault("represent_as", 0, defaultValue=None)
            #
            # ------- add the canonical identifiers --------
            cN = "rcsb_chem_comp_container_identifiers"
            if not dataContainer.exists(cN):
                dataContainer.append(DataCategory(cN, attributeNameList=self.__dApi.getAttributeNameList(cN)))
            idObj = dataContainer.getObj(cN)
            idObj.setValue(ccId, "comp_id", 0)
            if prdId:
                idObj.setValue(prdId, "prd_id", 0)
            idObj.setValue(ccId, "rcsb_id", 0)
            if subComponentIds:
                tL = [tV.strip() for tV in subComponentIds.split()]
                idObj.setValue(",".join(tL), "subcomponent_ids", 0)
            #
            # Get audit info -
            if representAs and representAs.lower() in ["polymer"]:
                _, releaseDate, revisionDate = self.__getAuditDates(dataContainer, "pdbx_prd_audit")
            else:
                _, releaseDate, revisionDate = self.__getAuditDates(dataContainer, "pdbx_chem_comp_audit")
            #
            #  --------- --------- --------- ---------
            # Create the new target category
            #
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            #
            # -------
            wObj = dataContainer.getObj(catName)
            #
            numAtoms = 0
            numAtomsHeavy = 0
            numAtomsChiral = 0
            try:
                cObj = dataContainer.getObj("chem_comp_atom")
                numAtoms = cObj.getRowCount()
                numAtomsHeavy = 0
                numAtomsChiral = 0
                for ii in range(numAtoms):
                    el = cObj.getValue("type_symbol", ii)
                    if el != "H":
                        numAtomsHeavy += 1
                    chFlag = cObj.getValue("pdbx_stereo_config", ii)
                    if chFlag != "N":
                        numAtomsChiral += 1
            except Exception:
                logger.warning("Missing chem_comp_atom category for %s", ccId)
                numAtoms = 0
                numAtomsHeavy = 0
                numAtomsChiral = 0
            #
            wObj.setValue(ccId, "comp_id", 0)
            if prdReleaseStatus:
                wObj.setValue(prdReleaseStatus, "release_status", 0)
            else:
                wObj.setValue(ccReleaseStatus, "release_status", 0)
            #
            wObj.setValue(releaseDate, "initial_release_date", 0)
            wObj.setValue(revisionDate, "revision_date", 0)
            #
            wObj.setValue(numAtoms, "atom_count", 0)
            wObj.setValue(numAtomsChiral, "atom_count_chiral", 0)
            wObj.setValue(numAtomsHeavy, "atom_count_heavy", 0)
            #
            #  ------
            numBonds = 0
            numBondsAro = 0
            try:
                cObj = dataContainer.getObj("chem_comp_bond")
                numBonds = cObj.getRowCount()
                numBondsAro = 0
                for ii in range(numAtoms):
                    aroFlag = cObj.getValue("pdbx_aromatic_flag", ii)
                    if aroFlag != "N":
                        numBondsAro += 1
            except Exception:
                pass
            #
            wObj.setValue(numBonds, "bond_count", 0)
            wObj.setValue(numBondsAro, "bond_count_aromatic", 0)
            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def addChemCompDescriptor(self, dataContainer, catName, **kwargs):
        """Add category rcsb_chem_comp_descriptor.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example, parse the pdbx_chem_comp_descriptor category and extract SMILES/CACTVS and InChI descriptors -

        loop_
        _pdbx_chem_comp_descriptor.comp_id
        _pdbx_chem_comp_descriptor.type
        _pdbx_chem_comp_descriptor.program
        _pdbx_chem_comp_descriptor.program_version
        _pdbx_chem_comp_descriptor.descriptor
        ATP SMILES           ACDLabs              10.04 "O=P(O)(O)OP(=O)(O)OP(=O)(O)OCC3OC(n2cnc1c(ncnc12)N)C(O)C3O"
        ATP SMILES_CANONICAL CACTVS               3.341 "Nc1ncnc2n(cnc12)[C@@H]3O[C@H](CO[P@](O)(=O)O[P@@](O)(=O)O[P](O)(O)=O)[C@@H](O)[C@H]3O"
        ATP SMILES           CACTVS               3.341 "Nc1ncnc2n(cnc12)[CH]3O[CH](CO[P](O)(=O)O[P](O)(=O)O[P](O)(O)=O)[CH](O)[CH]3O"
        ATP SMILES_CANONICAL "OpenEye OEToolkits" 1.5.0 "c1nc(c2c(n1)n(cn2)[C@H]3[C@@H]([C@@H]([C@H](O3)CO[P@@](=O)(O)O[P@](=O)(O)OP(=O)(O)O)O)O)N"
        ATP SMILES           "OpenEye OEToolkits" 1.5.0 "c1nc(c2c(n1)n(cn2)C3C(C(C(O3)COP(=O)(O)OP(=O)(O)OP(=O)(O)O)O)O)N"
        ATP InChI            InChI                1.03  "InChI=1S/C10H16N5O13P3/c11-8-5-9(13-2-12-8)15(3- ...."
        ATP InChIKey         InChI                1.03  ZKHQWZAMYRWXGA-KQYNXXCUSA-N

        To produce -
             _rcsb_chem_comp_descriptor.comp_id                 ATP
             _rcsb_chem_comp_descriptor.SMILES                  'Nc1ncnc2n(cnc12)[CH]3O[CH](CO[P](O)(=O)O[P](O)(=O)O[P](O)(O)=O)[CH](O)[CH]3O'
             _rcsb_chem_comp_descriptor.SMILES_stereo           'Nc1ncnc2n(cnc12)[C@@H]3O[C@H](CO[P@](O)(=O)O[P@@](O)(=O)O[P](O)(O)=O)[C@@H](O)[C@H]3O'
             _rcsb_chem_comp_descriptor.InChI                   'InChI=1S/C10H16N5O13P3/c11-8-5-9(13-2-12-8)15(3-14-5)10-7(17)6(16)4(26-10)1-25 ...'
             _rcsb_chem_comp_descriptor.InChIKey                'ZKHQWZAMYRWXGA-KQYNXXCUSA-N'
        """
        try:
            logger.debug("Starting with  %r %r %r", dataContainer.getName(), catName, kwargs)
            # Exit if source categories are missing
            if not (dataContainer.exists("chem_comp") and dataContainer.exists("pdbx_chem_comp_descriptor")):
                return False
            #
            # Create the new target category
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["comp_id", "SMILES", "SMILES_stereo", "InChI", "InChIKey"]))
            #
            wObj = dataContainer.getObj(catName)
            ccIObj = dataContainer.getObj("pdbx_chem_comp_descriptor")
            iRow = 0
            ccId = ""
            for ii in range(ccIObj.getRowCount()):
                ccId = ccIObj.getValue("comp_id", ii)
                nm = ccIObj.getValue("descriptor", ii)
                prog = ccIObj.getValue("program", ii)
                typ = ccIObj.getValue("type", ii)
                #
                if typ == "SMILES_CANONICAL" and prog == "CACTVS":
                    wObj.setValue(nm, "SMILES_stereo", iRow)
                elif typ == "SMILES" and prog == "CACTVS":
                    wObj.setValue(nm, "SMILES", iRow)
                elif typ == "InChI" and prog == "InChI":
                    wObj.setValue(nm, "InChI", iRow)
                elif typ == "InChIKey" and prog == "InChI":
                    wObj.setValue(nm, "InChIKey", iRow)
            #
            wObj.setValue(ccId, "comp_id", iRow)
            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def renameCitationCategory(self, dataContainer, catName, **kwargs):
        """Rename citation and citation author categories.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise
        """
        try:
            _ = kwargs
            logger.debug("Starting with  %r %r", dataContainer.getName(), catName)
            if not (dataContainer.exists("chem_comp") and dataContainer.exists("pdbx_chem_comp_identifier")):
                return False
            #
            # Rename target categories
            if dataContainer.exists("citation"):
                dataContainer.rename("citation", "rcsb_bird_citation")
            if dataContainer.exists("citation_author"):
                dataContainer.rename("citation_author", "rcsb_bird_citation_author")
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))

        return False

    def addChemCompSynonyms(self, dataContainer, catName, **kwargs):
        """Add category rcsb_chem_comp_synonyms including PDB and DrugBank annotations.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example,

             loop_
                 _rcsb_chem_comp_synonyms.comp_id
                 _rcsb_chem_comp_synonyms.ordinal
                 _rcsb_chem_comp_synonyms.name
                 _rcsb_chem_comp_synonyms.provenance_code
                    ATP 1 "adenosine 5'-(tetrahydrogen triphosphate)"  'PDB Reference Data'
                    ATP 2 "Adenosine 5'-triphosphate"  'PDB Reference Data'
                    ATP 3 Atriphos  DrugBank
                    ATP 4 Striadyne DrugBank

        """
        try:
            logger.debug("Starting with  %r %r", dataContainer.getName(), catName)
            if not (dataContainer.exists("chem_comp") and dataContainer.exists("pdbx_chem_comp_identifier")):
                return False
            #
            #
            # Create the new target category
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["comp_id", "ordinal", "name", "provenance_source"]))
            else:
                # remove the rowlist -
                pass
            #
            pTupL = self.__dApi.getEnumListWithDetail(catName, "provenance_source")
            provLookupD = {pTup[0].upper(): pTup[0] for pTup in pTupL}
            provLookupD["ACD-LABS"] = "ACDLabs"
            provLookupD["PDB"] = "PDB Reference Data"

            wObj = dataContainer.getObj(catName)
            #
            # Get all of the names relevant names from the definition -
            #
            iRow = 0
            provCode = "PDB Reference Data"
            ccObj = dataContainer.getObj("chem_comp")
            ccId = ccObj.getValue("id", 0)
            ccName = ccObj.getValue("name", 0)
            ccSynonymL = []
            if ccObj.hasAttribute("pdbx_synonyms"):
                ccSynonymL = str(ccObj.getValue("pdbx_synonyms", 0)).split(";")
            #
            wObj.setValue(ccId, "comp_id", iRow)
            wObj.setValue(ccName, "name", iRow)
            wObj.setValue(iRow + 1, "ordinal", iRow)
            wObj.setValue(provCode, "provenance_source", iRow)
            iRow += 1
            #
            if dataContainer.exists("pdbx_chem_comp_synoynms"):
                qObj = dataContainer.getObj("pdbx_chem_comp_synoynms")
                for ii in range(qObj.getRowCount()):
                    nm = qObj.getValue("name", ii)
                    wObj.setValue(ccId, "comp_id", iRow)
                    wObj.setValue(nm, "name", iRow)
                    wObj.setValue(iRow + 1, "ordinal", iRow)
                    wObj.setValue(provCode, "provenance_source", iRow)
                    iRow += 1

            for nm in ccSynonymL:
                if nm in ["?", "."]:
                    continue
                wObj.setValue(ccId, "comp_id", iRow)
                wObj.setValue(nm, "name", iRow)
                wObj.setValue(iRow + 1, "ordinal", iRow)
                wObj.setValue(provCode, "provenance_source", iRow)
                iRow += 1
            #
            ccIObj = dataContainer.getObj("pdbx_chem_comp_identifier")
            for ii in range(ccIObj.getRowCount()):
                nm = ccIObj.getValue("identifier", ii)
                prog = ccIObj.getValue("program", ii)
                if prog and prog.upper() in provLookupD:
                    sProg = provLookupD[prog.upper()]
                    wObj.setValue(ccId, "comp_id", iRow)
                    wObj.setValue(nm, "name", iRow)
                    wObj.setValue(iRow + 1, "ordinal", iRow)
                    wObj.setValue(sProg, "provenance_source", iRow)
                    iRow += 1
                else:
                    logger.error("%s unknown provenance %r", ccId, prog)
            #
            rP = kwargs.get("resourceProvider")
            dbProvider = rP.getResource("DrugBankProvider instance") if rP else None
            dbD = dbProvider.getMapping()
            if dbD:
                dbMapD = dbD["id_map"]
                #
                if dbMapD and ccId in dbMapD and "aliases" in dbMapD[ccId]:
                    iRow = wObj.getRowCount()
                    for nm in dbMapD[ccId]["aliases"]:
                        wObj.setValue(ccId, "comp_id", iRow)
                        wObj.setValue(nm, "name", iRow)
                        wObj.setValue(iRow + 1, "ordinal", iRow)
                        wObj.setValue("DrugBank", "provenance_source", iRow)
                        iRow += 1

            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))

        return False
